
// ==========================================
//
// F6FBB 09/97 Portage DxNet F5MZN sous LINUX
//
// ==========================================


#include <stdio.h>
#define __USE_BSD
#include <string.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>

#include <sys/ioctl.h>
#include <sys/socket.h>
#include <net/if.h>
#include <linux/if_ether.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <linux/ax25.h>
#include <linux/netrom.h>
#include <linux/rose.h>

#include <bios.h>

#include "axutils.h"
#include "axconfig.h"
#include "nrconfig.h"
#include "rsconfig.h"
#include "alloc.h"

#include "define.h"
#include "node.h"
#include "tools.h"
#include "buffers.h"
#include "streams.h"
#include "password.h"
#include "users.h"
#include "filter.h"
#include "switch.h"
#include "commands.h"
#include "set.h"

typedef struct _OrbClient
{
	int fd;
	int state;
	int monitor;
	int channel;
	long mask;
	time_t cle;
	char callsign[12];
	struct _OrbClient *next;
}
OrbClient;

static int orb_fd = -1;
static OrbClient * client_head = NULL;
static int IsConsoleConnected = FALSE;

#define AX25_CALLSID 11			/* Big enough for ROSE addresses */

#define READ_EVENT 1
#define WRITE_EVENT 2
#define EXCEPT_EVENT 4

#define DISCONNECT 0
#define FAILURE    1
#define CONNECT    2
#define CPROGRESS  3
#define WAITCALL   4
#define WAITPASS   5

#define NO_EVENT   0
#define DISC_EVENT 1
#define CONN_EVENT 2
#define RETR_EVENT 3
#define BUSY_EVENT 4

#define SOCK_MAXCAN 50

#define	CAN_AX25	0
#define	CAN_NETROM	1
#define	CAN_ROSE	2
#define	CAN_INET	3

#define ORB_WAITINFO	0
#define ORB_WAITPASS	1
#define ORB_CONNECTED	2
#define ORB_DISCONNECT	3

#define ORB_MSGS	1
#define ORB_STATUS	2
#define ORB_NBCNX	4
#define ORB_LISTCNX	8
#define ORB_MONITOR	16
#define ORB_CONSOLE	32
#define ORB_CHANNEL	64
#define ORB_XFBBX	128

#define W_CNST		5

char MyCall[20] = "NOCALL";
char *TelnetInit = "\377\374\001\000Telnet Initialisation string";
char *Hello = "%s\r%s DxNet. TELNET access\r\r%s";
char *Callsign = "Callsign : ";
char *Password = "Password : ";
char *Logon = "\rLogon Ok\r\r";

int FirstStream = 1;
int LastStream = NBSTREAMS;

int SWITCH_nSocket = 0;

union sockaddr_ham
{
	struct full_sockaddr_ax25 axaddr;
	struct sockaddr_rose rsaddr;
	struct sockaddr_in inaddr;
};

typedef struct Port
{
	char Name[80];
	int Type;
	int Sock;
	struct Port *Next;
}
PORT;

typedef struct
{
	int fd;						// fd du socket connecte

	int Type;					// Type du port connecte

	int Queue;					// Longueur de la queue

	int Echec;					// Nbre d'echecs password

	char CallSign[20];			// indicatif connecte

	char State;					// Etat de la connexion

	char Event;					// Evenement recu

	char Delta;					// L'etat a change
	
	int Paclen;					//Modif F5MZN 07/02/1999 : Gestion du Paclen
	
	char Port[20];

}
QSO;

QSO QsoTab[NBSTREAMS + 1];		// Reserver de la memoire pour 65 QSO

PORT *PortHead = NULL;			// Liste des ports a ecouter

extern char STREAMS_level[MAX_STREAMS];

//
// Fonctions statiques
//
static int s_free (int);
static int s_status (int);
static int sock_connexion (int, int, union sockaddr_ham *);
static int rs_convert_call (char *address, struct sockaddr_rose *addr);
static int nr_convert_call (char *address, struct full_sockaddr_ax25 *addr);
static int in_convert_call (char *address, struct sockaddr_in *addr, char *Callsign);
static int telnet_check (int StreamNum, char *buffer, int ilen);

static char *ax25_ntoa (char *, const ax25_address *);
static char *reason (unsigned char cause, unsigned char diagnostic);

/* Une seule initialisation quelque soit le driver */
static int fbb_ax25_config_load_ports ()
{
	static int init_ok = 0;

	if (init_ok)
		return (1);

	init_ok = 1;
	return (ax25_config_load_ports ());
}

/* Une seule initialisation quelque soit le driver */
static int fbb_nr_config_load_ports ()
{
	static int init_ok = 0;

	if (init_ok)
		return (1);

	init_ok = 1;
	return (nr_config_load_ports ());
}

/* Une seule initialisation quelque soit le driver */
static int fbb_rs_config_load_ports ()
{
	static int init_ok = 0;

	if (init_ok)
		return (1);

	init_ok = 1;
	return (rs_config_load_ports ());
}

static void orb_send_data (char *buffer, int lg, int mask)
{
	OrbClient *sptr = client_head;

	while (sptr)
	{
		if (sptr->mask & mask)
		{
			switch (mask)
			{
			case ORB_MONITOR:
			case ORB_CONSOLE:
			case ORB_CHANNEL:
				if (((mask == ORB_MONITOR) && (buffer[4] == 0xff)) ||
					((mask == ORB_CONSOLE) && (buffer[4] == 0)))
				{
					write (sptr->fd, buffer, lg);
				}
				if ((mask == ORB_CHANNEL) && (buffer[4] > 0) && (buffer[4] < 0xff))
				{
					if ((sptr->channel == 0) || (sptr->channel == buffer[4]))
						write (sptr->fd, buffer, lg);
				}
				break;
			default:
				write (sptr->fd, buffer, lg);
				break;
			}
		}
		sptr = sptr->next;
	}
}

/*----------------------------------------------------------------------------
  ------ Envoyer des datas a la console --------------------------------------
  ----------------------------------------------------------------------------
  Extrait de fbb_orb.c (f6fbb)                                                 
*/
void SWITCH_orbWrite (int channel, char *data, int len, int color, int header)
{
	int i, j;
	char *buffer;

	buffer = (char*) malloc (len + 8);
	if (buffer == NULL)
		return;

	for (i = 0, j = 7; i < len; i++)
	{
		if (data[i] == '\n')
			continue;

		if (data[i] == '\r')
			buffer[j++] = '\n';
		else
			buffer[j++] = data[i];
	}

	if (channel == _CONSOLE)
	{
		/* Header */
		buffer[0] = ORB_CONSOLE;
		buffer[1] = 0;
		buffer[2] = (j - 4) & 0xff;
		buffer[3] = (j - 4) >> 8;

		/* Envoie les data de console au client */
		buffer[4] = _CONSOLE;
		buffer[5] = color;
		buffer[6] = header;
		orb_send_data (buffer, j, ORB_CONSOLE);
	}
	else if (channel == _MONITOR)
	{
		/* Header */
		buffer[0] = ORB_MONITOR;
		buffer[1] = 0;
		buffer[2] = (j - 4) & 0xff;
		buffer[3] = (j - 4) >> 8;

		/* Envoie les data de monitoring au client */
		buffer[4] = 0xff;
		buffer[5] = color;
		buffer[6] = header;
		orb_send_data (buffer, j, ORB_MONITOR);
	}
	else
	{
		/* Header */
		buffer[0] = ORB_CHANNEL;
		buffer[1] = 0;
		buffer[2] = (j - 4) & 0xff;
		buffer[3] = (j - 4) >> 8;

		/* Envoie les data de monitoring au client */
		buffer[4] = channel - 1;
		buffer[5] = color;
		buffer[6] = header;
		orb_send_data (buffer, j, ORB_CHANNEL);
	}
	free (buffer);
}

/*----------------------------------------------------------------------------
  ------ Initialisation du socket destine aux consoles -----------------------
  ----------------------------------------------------------------------------
  Extrait de fbb_orb.c (f6fbb)                                                 
  Retour : 1 si OK
*/
int SWITCH_initOrb (int port)
{
	int val;
	int len;
	struct sockaddr_in sock_addr;

	if ((orb_fd = socket (AF_INET, SOCK_STREAM, 0)) < 0)
	{
		perror ("socket_r");
		return (0);
	}

	val = 1;
	len = sizeof (val);
	if (setsockopt (orb_fd, SOL_SOCKET, SO_REUSEADDR, (char *) &val, len) == -1)
	{
		perror ("opn_tcp : setsockopt SO_REUSEADDR");
	}

	sock_addr.sin_family = AF_INET;
	sock_addr.sin_addr.s_addr = 0;
	sock_addr.sin_port = htons (port);

	if (bind (orb_fd, (struct sockaddr *) &sock_addr, sizeof (sock_addr)) != 0)
	{
		perror ("opn_tcp : bind");
		close (orb_fd);
		return (0);
	}

	if (listen (orb_fd, SOMAXCONN) == -1)
	{
		perror ("listen");
		close (orb_fd);
		return (0);
	}
	
	printf("xfbbC/X server running on port %d ...\n", port);

	return (1);
}

/*----------------------------------------------------------------------------
  ------ Ajouter un nouveau client a la liste chainee ------------------------
  ----------------------------------------------------------------------------
  Extrait de fbb_orb.c (f6fbb)                                                
  retour = pointeur sur la nouvelle structure OrbClient ou NULL si echec
*/   
static OrbClient *orb_add_client (void)
{
	OrbClient *sptr;

	sptr = (OrbClient*) calloc (1, sizeof (OrbClient));
	if (sptr == NULL)
		return (NULL);

	sptr->next = client_head;
	client_head = sptr;

	return (sptr);
}

/*----------------------------------------------------------------------------
  ------ Supprimer un client de la liste chainee -----------------------------
  ----------------------------------------------------------------------------
  Extrait de fbb_orb.c (f6fbb)                                                
*/   
static void orb_del_client (OrbClient * cptr)
{
	OrbClient *sptr;
	OrbClient *prev;

	prev = NULL;
	sptr = client_head;
	while (sptr)
	{
		if (sptr == cptr)
		{
			if (prev)
				prev->next = sptr->next;
			else
				client_head = sptr->next;
			free (sptr);
			break;
		}
		prev = sptr;
		sptr = sptr->next;
	}
}

/*----------------------------------------------------------------------------
  ------ Accepter une nouvelle connexion d'un client -------------------------
  ----------------------------------------------------------------------------
  Extrait de fbb_orb.c (f6fbb)                                                
  retour = 1 si OK
*/   
static int orb_new_connection (int fd)
{
	int addr_len;
	struct sockaddr_in sock_addr;
	OrbClient *sptr;

	sptr = orb_add_client ();

	addr_len = sizeof (sock_addr);

	sptr->fd = accept (fd, (struct sockaddr *) &sock_addr, &addr_len);
	sptr->state = ORB_WAITINFO;

	return (sptr->fd);
}


/*---------------------------------------------------------------------------
 ------ Tente de connecter la console ---------------------------------------
 ----------------------------------------------------------------------------
 Retour : TRUE si OK
*/
static int console_connect (OrbClient * sptr)
{
	int nFlags;					
	extern char STREAMS_callsign[MAX_STREAMS][10];
	
	if( IsConsoleConnected )
		return FALSE;
					
	IsConsoleConnected = TRUE;
				
	strcpy(STREAMS_callsign[BUFFER_SCREEN2], sptr->callsign);
					
	/* Charger la config utilisateur pour l'indicatif */
	USERS_openRecord(BUFFER_SCREEN2, sptr->callsign);

	/* Initialiser les filtres */
	nFlags = USERS_getFlags(BUFFER_SCREEN2);
	FILTER_setUser(BUFFER_SCREEN2, nFlags);
					
	/* Envoyer le CTEXT */
	STREAMS_ctext(BUFFER_SCREEN2);
	
	return TRUE;
}

/*---------------------------------------------------------------------------
 ------ Tente de deconnecter la console -------------------------------------
 ----------------------------------------------------------------------------
 Retour : TRUE si OK
*/
static void console_disconnect (void)
{
	int nFlags;					
	extern char STREAMS_callsign[MAX_STREAMS][10];

	IsConsoleConnected = FALSE;
				
	strcpy(STREAMS_callsign[BUFFER_SCREEN2], MyCall);
					
	/* Charger la config utilisateur pour l'indicatif */
	USERS_openRecord(BUFFER_SCREEN2, MyCall);

	/* Initialiser les filtres */
	nFlags = USERS_getFlags(BUFFER_SCREEN2);
	FILTER_setUser(BUFFER_SCREEN2, nFlags);
}

static void orb_process_data (OrbClient * sptr)
{
	extern char * PARAMS_sysop;
	int nb;
	static char buffer[1024];
	static int nbrcv = 0;
	char call[256];
	char pass[256];
	char *ptr;
	int ok = FALSE;	

	nb = read (sptr->fd, buffer + nbrcv, sizeof (buffer) - nbrcv);
	if (nb == -1)
		return;

	if (nb == 0)
	{
		/* Client is disconnected */
		if (sptr->mask & ORB_CONSOLE)
		{
			/* Disconnect from console */
			console_disconnect ();
		}

		close (sptr->fd);
		orb_del_client (sptr);
		nbrcv = 0;
		return;
	}

	nbrcv += nb;

	if ((nbrcv == 3) && (buffer[0] == '\0'))
	{
		int mask;
		int old_mask = sptr->mask;
		int ack;

		/*commande */
		switch (buffer[1])
		{
		case 0:
			/* Changement de masque */
			mask = buffer[2] & 0xff;
			if ((mask & ORB_CONSOLE) != (sptr->mask & ORB_CONSOLE))
			{
				if (mask & ORB_CONSOLE)
				{
					/* Console connection */
					if( console_connect (sptr) == TRUE )
					{
						ack = 1;
					}
					else
					{
						ack = 2;
					}
				}
				else
				{
					/* Console deconnection */
					console_disconnect ();
					ack = 0;
				}
				buffer[0] = ORB_XFBBX;
				buffer[1] = 0;
				buffer[2] = 1;
				buffer[3] = 0;
				buffer[4] = ack;
				write (sptr->fd, buffer, 5);
			}
			sptr->mask = mask;
		}
		nbrcv = 0;
		return;
	}

	/* Check if the whole line is received */
	buffer[nbrcv] = '\0';
	if (strchr (buffer, '\n') == NULL)
		return;

	switch (sptr->state)
	{
	case ORB_WAITINFO:

		/* Mask of services */
		sscanf (buffer, "%ld %d %s", &sptr->mask, &sptr->channel, call);
		call[11] = '\0';
		strcpy (sptr->callsign, strupr(call));
		sptr->cle = time (NULL);
		sprintf (call, "%010ld", sptr->cle);
		write (sptr->fd, call, strlen (call));
		sptr->state = ORB_WAITPASS;
		break;

	case ORB_WAITPASS:
	
		/* L'indicatif est-il declare sysop ?*/
  		strcpy(call, sptr->callsign);

  		/* Chercher d'abord pour un SSID donne*/
  		if( ! strchr(call, '-') )
    			strcat(call, "-0"); /* Ajouter le SSID -0 si besoin */
  		if( SET_check(PARAMS_sysop, call) )
    			ok = TRUE;

  		/* Chercher maintenant sans tenir compte du SSID */
  		TOOLS_removeSsid(call);
  		if( SET_check(PARAMS_sysop, call) )
    			ok = TRUE;	

		/* Callsign sans SSID */
		strcpy (call, sptr->callsign);
		if ((ptr = strchr (call, '-')))
			*ptr = '\0';

		/* Password */
		sscanf (buffer, "%s", pass);

		if( ok && PASSWORD_isMDKeyOK(call, pass, sptr->cle) )
		{
			if (sptr->mask & ORB_CONSOLE)
			{
				if( console_connect (sptr) == FALSE )
				{
					char *txt = "Console already connected !\n\n";
					int len = strlen (txt) + 3;

					strcpy (buffer + 7, txt);

					/* Header */
					buffer[0] = ORB_CONSOLE;
					buffer[1] = 0;
					buffer[2] = (len) & 0xff;
					buffer[3] = (len) >> 8;

					buffer[4] = ORB_CONSOLE;
					buffer[5] = W_CNST;
					buffer[6] = 0;

					write (sptr->fd, buffer, len + 4);
					close (sptr->fd);
					orb_del_client (sptr);
				}
			}
			sptr->state = ORB_CONNECTED;
		}
		else
		{
			char *txt = "Callsign/Password error !\n\n";
			int len = strlen (txt) + 3;

			strcpy (buffer + 7, txt);

			/* Header */
			buffer[0] = ORB_CONSOLE;
			buffer[1] = 0;
			buffer[2] = (len) & 0xff;
			buffer[3] = (len) >> 8;

			buffer[4] = ORB_CONSOLE;
			buffer[5] = W_CNST;
			buffer[6] = 0;

			write (sptr->fd, buffer, len + 4);
			close (sptr->fd);
			orb_del_client (sptr);
		}
		break;

	case ORB_CONNECTED:

		if (sptr->mask & ORB_CONSOLE)
		{
			TOOLS_removeNR(buffer);
			CMD_execute(BUFFER_SCREEN2, buffer);
		}

		break;
	}

	nbrcv = 0;
}

static int SWITCH_configPort (char *MyCall, PORT * Port)
{
	char *p_name;
	int addrlen;
	int val;
	int tcp_port;
	union sockaddr_ham addr;

	if (strcmp ("*", Port->Name) == 0)
	{
		Port->Type = CAN_AX25;
	}
	else if (strncmp ("inet", Port->Name, 4) == 0)
	{
		Port->Type = CAN_INET;
	}
	else if (ax25_config_get_addr (Port->Name) != NULL)
	{
		Port->Type = CAN_AX25;
	}
	else if (nr_config_get_addr (Port->Name) != NULL)
	{
		Port->Type = CAN_NETROM;
	}
	else if (rs_config_get_addr (Port->Name) != NULL)
	{
		Port->Type = CAN_ROSE;
	}
	else
	{
		fprintf (stderr, "Invalid port name %s\n", Port->Name);
		return (-1);
	}

	switch (Port->Type)
	{

	case CAN_AX25:
		// Socket sur le port en attente de connexions
		if ((Port->Sock = socket (AF_AX25, SOCK_SEQPACKET, 0)) < 0)
		{
			perror ("socket_r");
			return -1;
		}

		addr.axaddr.fsa_ax25.sax25_family = AF_AX25;
		addr.axaddr.fsa_ax25.sax25_ndigis = 1;
		convert_call_entry (MyCall, addr.axaddr.fsa_ax25.sax25_call.ax25_call);
		if (strcmp ("*", Port->Name) == 0)
		{
			addr.axaddr.fsa_digipeater[0] = null_ax25_address;
		}
		else
		{
			p_name = ax25_config_get_addr (Port->Name);
			if (p_name == NULL)
				return -1;
			convert_call_entry (p_name, addr.axaddr.fsa_digipeater[0].ax25_call);
		}
		addrlen = sizeof (struct full_sockaddr_ax25);

		break;

	case CAN_NETROM:
		if ((Port->Sock = socket (AF_NETROM, SOCK_SEQPACKET, 0)) < 0)
		{
			perror ("socket_r");
			return -1;
		}

		addr.axaddr.fsa_ax25.sax25_family = AF_NETROM;
		addr.axaddr.fsa_ax25.sax25_ndigis = 0;
		p_name = nr_config_get_addr (Port->Name);
		convert_call_entry (p_name, addr.axaddr.fsa_ax25.sax25_call.ax25_call);
		addrlen = sizeof (struct full_sockaddr_ax25);

		break;

	case CAN_ROSE:
		if ((Port->Sock = socket (AF_ROSE, SOCK_SEQPACKET, 0)) < 0)
		{
			perror ("socket_r");
			return -1;
		}

		addr.rsaddr.srose_family = AF_ROSE;
		addr.rsaddr.srose_ndigis = 0;
		convert_call_entry (MyCall, addr.rsaddr.srose_call.ax25_call);
		p_name = rs_config_get_addr (Port->Name);
		convert_rose_address (p_name, addr.rsaddr.srose_addr.rose_addr);
		addrlen = sizeof (struct sockaddr_rose);

		break;

	case CAN_INET:
		if ((Port->Sock = socket (AF_INET, SOCK_STREAM, 0)) < 0)
		{
			perror ("socket_r");
			return -1;
		}

		val = 1;
		addrlen = sizeof (val);
		if (setsockopt (Port->Sock, SOL_SOCKET, SO_REUSEADDR, (char *) &val, addrlen) == -1)
		{
			perror ("opn_tcp : setsockopt SO_REUSEADDR");
		}

		p_name = strrchr (Port->Name, ':');
		if (p_name)
			tcp_port = atoi (p_name + 1);
		else
			tcp_port = 23;
		addr.inaddr.sin_family = AF_INET;
		addr.inaddr.sin_addr.s_addr = 0;
		addr.inaddr.sin_port = htons (tcp_port);

		addrlen = sizeof (struct sockaddr_in);

		break;
	}

	if (bind (Port->Sock, (struct sockaddr *) &addr, addrlen) == -1)
	{
		perror ("bind");
		close (Port->Sock);
		Port->Sock = -1;
		return -1;
	}

	if (listen (Port->Sock, SOMAXCONN) == -1)
	{
		perror ("listen");
		close (Port->Sock);
		Port->Sock = -1;
		return -1;
	}

	return (Port->Sock);
}

// Initialise ax25
int SWITCH_initSwitch (int, int, int)
{
	int i;
	PORT *Port;
	static int init_ok = 0;

	if (init_ok)
	{
		for (Port = PortHead; Port; Port = Port->Next)
		{
			if (Port->Sock >= 0)
				close (Port->Sock);
			Port->Sock = 0;
		}
	}
	else
	{
		// Initialiser les structures QSO
		for (i = 0; i < NBSTREAMS + 1; i++)
			memset (&QsoTab[i], 0, sizeof (QSO));

		fbb_ax25_config_load_ports ();
		fbb_nr_config_load_ports ();
		fbb_rs_config_load_ports ();

		// Recupere l'indicatif par default du DxNet
		NODE_getNodeCall (0, 0, MyCall);

		printf ("Found ax25 kernel.\n");
	}

	if (PortHead == NULL)
	{
		PortHead = (PORT *) calloc (sizeof (PORT), 1);
		strcpy (PortHead->Name, "*");
		PortHead->Sock = -1;
	}
	for (Port = PortHead; Port; Port = Port->Next)
	{
		SWITCH_configPort (MyCall, Port);
	}

	init_ok = 1;

	return TRUE;
}

void SWITCH_addPort (char *PortName)
{
	PORT *Port;

	if ((PortHead) && (strcmp (PortHead->Name, "*") == 0))
	{
		// Deallocates Default struct
		if (PortHead->Sock >= 0)
			close (PortHead->Sock);
		free (PortHead);
		PortHead = NULL;
	}

	Port = (PORT *) calloc (sizeof (PORT), 1);
	if (Port == NULL)
	{
		fprintf (stderr, "Cannot allocate PORT structure\n");
		return;
	}
	strcpy (Port->Name, PortName);
	Port->Sock = -1;
	Port->Next = PortHead;
	PortHead = Port;

	// Re-inits the switch
	if (*MyCall)
		SWITCH_initSwitch (0, 0, 0);
}


//---------------------------------------------------------------------------
//------ SetMyCall ----------------------------------------------------------
//---------------------------------------------------------------------------
void SWITCH_setMyCall (char *NodeCall)
{
	strcpy (MyCall, NodeCall);
	SWITCH_initSwitch (0, 0, 0);
}


//--------------------------------------------------------------------------
//------ wait - Waits until an event or timeout ----------------------------
//--------------------------------------------------------------------------
int SWITCH_wait (int delay)	//delay est en secondes
{
	int n;
	int i;
	int ret = 0;
	int last_fd = orb_fd;
	OrbClient * sptr;

	struct timeval timeval;
	fd_set sock_read;
	fd_set sock_excp;

	FD_ZERO (&sock_read);
	FD_ZERO (&sock_excp);

	// Keyboard ...
	if (stdin_valid)
		FD_SET (0, &sock_read);

	for (i = 0; i < NBSTREAMS + 1; i++)
	{
		if (QsoTab[i].fd > last_fd)
			last_fd = QsoTab[i].fd;
		if (QsoTab[i].fd)
		{
			FD_SET (QsoTab[i].fd, &sock_read);
			FD_SET (QsoTab[i].fd, &sock_excp);
		}
	}
	
	/* Clients console */
	if( orb_fd >= 0 )
	{
		FD_SET (orb_fd, &sock_read);
		sptr = client_head;
		while (sptr)
		{
			FD_SET(sptr->fd, &sock_read);
			if( sptr->fd > last_fd )
				last_fd = sptr->fd;
			sptr = sptr->next;
		}
	}	

//------- Modif F5MZN 6/6/99 --------
//	timeval.tv_usec = (delay % 1000) * 1000;
//	timeval.tv_sec = (delay / 1000);

	timeval.tv_usec = 100000;
	timeval.tv_sec  = 0;

	ret = select (last_fd + 1, &sock_read, NULL, &sock_excp, &timeval);

	if( ret <= 0 )
		return ret;	/* Rien a traiter */
		
	if (FD_ISSET (orb_fd, &sock_read))
	{
		/* Nouvelle connection client */
		orb_new_connection (orb_fd);
	}

	sptr = client_head;
	while (sptr)
	{
		/* On sauvegarde le next car le pointeur de client peut etre
		   libere pendant le traitement en cas de deconnexion */
		OrbClient *ptemp = sptr->next;

		if (FD_ISSET (sptr->fd, &sock_read))
		{
			orb_process_data (sptr);
		}
		sptr = ptemp;
	}

	return ret;
}

//--------------------------------------------------------------------------
//------ disc_switch - Dconnecte le stream du switch ----------------------
//--------------------------------------------------------------------------
void SWITCH_discSwitch (int StreamNum)
{
	if (QsoTab[StreamNum].fd)
	{
		close (QsoTab[StreamNum].fd);
		SWITCH_nSocket--;		
		QsoTab[StreamNum].fd = 0;
		printf ("*** Disconnected from %s\n", QsoTab[StreamNum].CallSign);
		memset (&QsoTab[StreamNum], 0, sizeof (QSO));
		QsoTab[StreamNum].Delta = 1;
		QsoTab[StreamNum].State = DISCONNECT;
	}
}

//--------------------------------------------------------------------------
//------ get_frame ---------------------------------------------------------
//--------------------------------------------------------------------------
//Lit et retourne dans buffer le contenu du cannal StreamNum
//La fonction retourne la longueur de la chaine
int SWITCH_getFrame (int StreamNum, char *buffer)
{
	int nb = 0;
	int res = s_status (StreamNum);

	if (res & EXCEPT_EVENT)
	{
		if (QsoTab[StreamNum].Event == CONN_EVENT)
		{
			/* Appel sortant connecte */
			unsigned int addr_len;
			union sockaddr_ham addr;
			int ret;

			addr_len = sizeof (union sockaddr_ham);

			if (getpeername (QsoTab[StreamNum].fd, (struct sockaddr *) &addr, (int*) &addr_len) == 0)
			{
				ret = sock_connexion (QsoTab[StreamNum].fd, StreamNum, &addr);
				if (ret)
				{
					return (0);
				}
			}
			else
				perror ("getpeername");
			return 0;
		}
		else
		{
			if (STREAMS_level[StreamNum] & LEVEL_linkSetup)
				BUFFERS_printBuff (StreamNum, IN, "*** No answer from %s\r", QsoTab[StreamNum].CallSign);
			printf ("*** No answer from %s\n", QsoTab[StreamNum].CallSign);
			if (QsoTab[StreamNum].Type == CAN_ROSE)
			{
				struct rose_cause_struct rose_cause;

				rose_cause.cause = ROSE_LOCAL_PROCEDURE;
				rose_cause.diagnostic = 0x12;

				ioctl (QsoTab[StreamNum].fd, SIOCRSGCAUSE, &rose_cause);
				printf ("*** %02X%02X - %s\n",
						rose_cause.cause,
						rose_cause.diagnostic,
						reason (rose_cause.cause, rose_cause.diagnostic));

			}
			close (QsoTab[StreamNum].fd);
			SWITCH_nSocket--;			
			QsoTab[StreamNum].fd = 0;
			QsoTab[StreamNum].Delta = TRUE;
			QsoTab[StreamNum].State = DISCONNECT;
			return 0;
		}
	}

	if (res & READ_EVENT)
	{
		if (QsoTab[StreamNum].fd <= 0)
		{
			printf ("read on invalide socket\n");
			return (0);
		}
		nb = read (QsoTab[StreamNum].fd, buffer, 256);
		if ((nb == 0) || ((nb == -1) && (errno == ENOTCONN)))
		{
			/* Deconnection */
			if (STREAMS_level[StreamNum] & LEVEL_linkSetup)
			{
				BUFFERS_printBuff (StreamNum, IN, "*** Disconnected from %s\r", QsoTab[StreamNum].CallSign);
			}
			printf ("*** Disconnected from %s\n", QsoTab[StreamNum].CallSign);
			if (QsoTab[StreamNum].Type == CAN_ROSE)
			{
				struct rose_cause_struct rose_cause;

				rose_cause.cause = ROSE_LOCAL_PROCEDURE;
				rose_cause.diagnostic = 0x12;

				ioctl (QsoTab[StreamNum].fd, SIOCRSGCAUSE, &rose_cause);
				printf ("*** %02X%02X - %s\n",
						rose_cause.cause,
						rose_cause.diagnostic,
						reason (rose_cause.cause, rose_cause.diagnostic));

			}
			close (QsoTab[StreamNum].fd);
			SWITCH_nSocket--;
			QsoTab[StreamNum].fd = 0;
			QsoTab[StreamNum].Delta = TRUE;
			QsoTab[StreamNum].State = DISCONNECT;
			nb = 0;
		}
		else if (nb == -1)
		{
			printf ("errno = %d\n", errno);
			perror ("read");
			nb = 0;
		}
		else
		{
			// Data available
			if (QsoTab[StreamNum].Type == CAN_INET)
			{
				int i;
				int ilen;
				char *pbuf = buffer;

				/* transforme crlf -> cr ... Pas de binaire ! */

				for (i = 0, ilen = 0; i < nb; ++i)
				{
					if ((buffer[i] == '\r') && (buffer[i + 1] == '\n'))
					{
						*pbuf++ = buffer[i++];
					}
					else
					{
						*pbuf++ = buffer[i];
					}
					ilen++;
				}
				nb = telnet_check (StreamNum, buffer, ilen);
			}
			buffer[nb] = '\0';
		}
	}
	if (res & WRITE_EVENT)
	{
		/* Can write to the socket... Unused */
	}
	
	return nb;
}

//--------------------------------------------------------------------------
//------ send_frame --------------------------------------------------------
//--------------------------------------------------------------------------
//Transmet le contenu de buffer dans le cannal StreamNum
void SWITCH_sendFrame (int StreamNum, char *buffer, int len)
{
	if (QsoTab[StreamNum].fd > 0)
	{
		int nb;

		if (QsoTab[StreamNum].Type == CAN_INET)
		{
			int i;
			int ilen;
			char *buf = (char *) malloc (len * 2);
			char *pbuf = buf;

			/* transforme cr -> crlf ... Pas de binaire ! */

			for (i = 0, ilen = 0; i < len; i++, ilen++)
			{
				*pbuf++ = *buffer;
				if (*buffer == '\r')
				{
					*pbuf++ = '\n';
					++ilen;
				}
				++buffer;
			}

			nb = write (QsoTab[StreamNum].fd, buf, ilen);

			free (buf);
			len = ilen;
		}
		else
		{
			nb = write (QsoTab[StreamNum].fd, buffer, len);
		}

		if (nb == -1)
		{
			printf ("Error %d on socket %d\n",
					errno, QsoTab[StreamNum].fd);
			perror ("write on socket");
		}
		else if (nb < len)
		{
			printf ("Cannot write %d bytes on socket %d\n",
					len - nb, QsoTab[StreamNum].fd);
		}
	}
}

//--------------------------------------------------------------------------
//------ class g8bpq - get_status ------------------------------------------
//--------------------------------------------------------------------------
//Retourne l'etat du canal : NONE pas de changement
//                           1 connection
//                           2 deconnection
//et initialise l'ack de test de status (#5) pour G8BPQ
int SWITCH_getStatut (int StreamNum)
{

	PORT *Port;

	// On en profilte pour tester s'il y a une connexion recue sur un des ports

	for (Port = PortHead; Port; Port = Port->Next)
	{

		QsoTab[0].fd = Port->Sock;

		int res = s_status (0);

		if (res & READ_EVENT)
		{
			int new_fd;
			int i;
			unsigned int addr_len;
			union sockaddr_ham addr;

			addr_len = sizeof (union sockaddr_ham);

			new_fd = accept (Port->Sock, (struct sockaddr *) &addr, (int*) &addr_len);
			SWITCH_nSocket++;

			/* Affect the new_fd socket to an empty channel */
			for (i = FirstStream; i <= LastStream; i++)
			{
				if (QsoTab[i].State == DISCONNECT)
				{
					break;
				}
			}

			if (i > LastStream)
			{
				/* Impossible d'affecter un stream -> deconnexion */
				printf ("no channel available\n");
				close (new_fd);
				SWITCH_nSocket--;
			}
			else
			{
				int ret;

				// Le stream i est affecte a la connexion entrante
				ret = sock_connexion (new_fd, i, &addr);
				
				switch (QsoTab[i].Type)		//F5MZN 07/02/1999
				{
				case CAN_AX25:
					QsoTab[i].Paclen = ax25_config_get_paclen (Port->Name);
					break;
				case CAN_NETROM:
					QsoTab[i].Paclen = nr_config_get_paclen (Port->Name);
					break;
				case CAN_ROSE:
					QsoTab[i].Paclen = rs_config_get_paclen (Port->Name);	
					break;
				case CAN_INET:
					QsoTab[i].Paclen = 256;	
					break;
				}
				
			}
		}
	}

	// Retourne l'etat du Stream demande
	if (QsoTab[StreamNum].Delta)
	{

		if ((QsoTab[StreamNum].State == DISCONNECT) ||
			(QsoTab[StreamNum].State == FAILURE))
		{
			// On nettoie la structure
			memset (&QsoTab[StreamNum], 0, sizeof (QSO));
			return DISCONNECTED;
		}
		else
		{
			// if (QsoTab[StreamNum].State == CPROGRESS)
			if (STREAMS_level[StreamNum] & LEVEL_linkSetup)
			{
				BUFFERS_printBuff (StreamNum, IN, "*** Connected to %s\r",
								   QsoTab[StreamNum].CallSign);
				QsoTab[StreamNum].State = CONNECT;
			}
			QsoTab[StreamNum].Delta = 0;
			return CONNECTED;
		}
	}

	//------ Pas de changement ------
	return NONE;
}

//--------------------------------------------------------------------------
//------ get_call ----------------------------------------------------------
//--------------------------------------------------------------------------
//Retourne dans call_sign l'indicatif connecte au cannal streamnul
void SWITCH_getCall (int StreamNum, char *call_sign)
{
	strcpy (call_sign, QsoTab[StreamNum].CallSign);
}

//--------------------------------------------------------------------------
//------ Monitoring --------------------------------------------------------
//--------------------------------------------------------------------------
//Monitoring
int SWITCH_rawRx (int StreamNum, char *buffer)
{
	//Ne fait rien pour l'instant
	return (0);
}

//---------------------------------------------------------------------------
//------ Is PcFlex ----------------------------------------------------------
//---------------------------------------------------------------------------
int SWITCH_isPcFlexNet (void)
{
	return 1;
}

//--------------------------------------------------------------------------
//------ con_state - Le stream est-il connect ? ---------------------------
//--------------------------------------------------------------------------
//Retourne 1 si le stream est connect
int SWITCH_conStat (int StreamNum)
{
	return (QsoTab[StreamNum].State != DISCONNECT);
}

//--------------------------------------------------------------------------
//------ ack_status - A appeler aprs chaque appel de int 5 ----------------
//--------------------------------------------------------------------------
//Cette fonction est utilise par SWITCH_getStat()
void SWITCH_ackStatut (int)
{
}

//--------------------------------------------------------------------------
//------ con_switch - Connecte le stream au switch -------------------------
//--------------------------------------------------------------------------
void SWITCH_conSwitch (int)
{
}

//--------------------------------------------------------------------------
//------ Connecter un indicatif via PcFlexNet ------------------------------
//--------------------------------------------------------------------------
//Retourne TRUE si le switch est un PcFlexNet
int SWITCH_connectFlex (int StreamNum, char *ToCall)
{
	int fd;
	int ioc;
	int type;
	int paclen;					//F5MZN 07/02/1999
	int addrlen;
	int bindlen;
	union sockaddr_ham caddr;	// Connect address

	union sockaddr_ham baddr;	// Bind address

	char *p_name;
	char *ptr;
	char str[256];

	strncpy (str, ToCall, 255);
	str[255] = '\0';

	TOOLS_removeNR (str);

	// Le premier mot doit etre un port valide
	ptr = strtok (str, " \t");
	if (ptr == NULL)
	{
		fprintf (stderr, "No specified port\n");
		return 0;
	}

	// Tester le type du port
	if (strcasecmp (ptr, "inet") == 0)
	{
		type = CAN_INET;
		paclen = 256;
	}
	else if (p_name = ax25_config_get_addr (ptr))
	{
		type = CAN_AX25;
		paclen = ax25_config_get_paclen (ptr);			//F5MZN 07/02/1999
	}
	else if (p_name = nr_config_get_addr (ptr))
	{
		type = CAN_NETROM;
		paclen = nr_config_get_paclen (ptr);			//F5MZN 07/02/1999
	}
	else if (p_name = rs_config_get_addr (ptr))
	{
		type = CAN_ROSE;
		paclen = rs_config_get_paclen (ptr);			//F5MZN 07/02/1999
	}
	else
	{
		fprintf (stderr, "Unknown port %s\n", ptr);
		return (0);
	}
	
	/* Nom du port */
	if( ptr && strlen(ptr) < 20 )
		strcpy(QsoTab[StreamNum].Port, ptr);
	else
		QsoTab[StreamNum].Port[0] = SNULL;


	// Destination appelee
	ptr = strtok (NULL, "");
	if (ptr == NULL)
	{
		fprintf (stderr, "No specified callsign\n");
		return 0;
	}

	switch (type)
	{

	case CAN_AX25:

		baddr.axaddr.fsa_ax25.sax25_family = AF_AX25;
		baddr.axaddr.fsa_ax25.sax25_ndigis = 1;

		// Indicatif appelant
		convert_call_entry (MyCall, baddr.axaddr.fsa_ax25.sax25_call.ax25_call);

		// Port
		convert_call_entry (p_name, baddr.axaddr.fsa_digipeater[0].ax25_call);

		// Indicatif appele et digis...
		if ((addrlen = convert_call (ptr, &caddr.axaddr)) == -1)
		{
			fprintf (stderr, "Unknown callsign %s\n", ptr);
			return 0;
		}

		strcpy (QsoTab[StreamNum].CallSign, ax2asc (&caddr.axaddr.fsa_ax25.sax25_call));

		bindlen = sizeof (struct full_sockaddr_ax25);

		if ((fd = socket (AF_AX25, SOCK_SEQPACKET, 0)) < 0)
		{
			perror ("socket_ax");
			return 0;
		}
		else
			SWITCH_nSocket++;

		break;

	case CAN_NETROM:

		baddr.axaddr.fsa_ax25.sax25_family = AF_NETROM;
		baddr.axaddr.fsa_ax25.sax25_ndigis = 1;

		// Indicatif appelant
		convert_call_entry (MyCall, baddr.axaddr.fsa_digipeater[0].ax25_call);

		// Port
		convert_call_entry (p_name, baddr.axaddr.fsa_ax25.sax25_call.ax25_call);

		// Indicatif appele et digis...
		if ((addrlen = nr_convert_call (ptr, &caddr.axaddr)) == -1)
		{
			fprintf (stderr, "Unknown callsign %s\n", ptr);
			return 0;
		}

		strcpy (QsoTab[StreamNum].CallSign, ax2asc (&caddr.axaddr.fsa_ax25.sax25_call));

		bindlen = sizeof (struct full_sockaddr_ax25);

		if ((fd = socket (AF_NETROM, SOCK_SEQPACKET, 0)) < 0)
		{
			perror ("socket_nr");
			return (0);
		}
		else
			SWITCH_nSocket++;

		break;

	case CAN_ROSE:

		baddr.rsaddr.srose_family = AF_ROSE;
		baddr.rsaddr.srose_ndigis = 0;

		// Indicatif appelant
		convert_call_entry (MyCall, baddr.rsaddr.srose_call.ax25_call);

		// Port
		convert_rose_address (p_name, baddr.rsaddr.srose_addr.rose_addr);

		// Indicatif appele, adresse et digis...
		if ((addrlen = rs_convert_call (ptr, &caddr.rsaddr)) == -1)
		{
			fprintf (stderr, "Unknown callsign %s\n", ptr);
			return 0;
		}

		strcpy (QsoTab[StreamNum].CallSign, ax2asc (&caddr.rsaddr.srose_call));

		bindlen = sizeof (struct sockaddr_rose);

		if ((fd = socket (AF_ROSE, SOCK_SEQPACKET, 0)) < 0)
		{
			perror ("socket_rs");
			return (0);
		}
		else
			SWITCH_nSocket++;

		break;

	case CAN_INET:

		// Indicatif appele et digis...
		if ((addrlen = in_convert_call (ptr, &caddr.inaddr, QsoTab[StreamNum].CallSign)) == -1)
		{
			fprintf (stderr, "Unknown callsign %s\n", ptr);
			return 0;
		}

		bindlen = 0;

		if ((fd = socket (AF_INET, SOCK_STREAM, 0)) < 0)
		{
			perror ("socket_ax");
			return 0;
		}
		else
			SWITCH_nSocket++;

		break;

	}

	if (bindlen && bind (fd, (struct sockaddr *) &baddr, bindlen) == -1)
	{
		perror ("sock_connect : bind");
		close (fd);
		SWITCH_nSocket--;
		free (str);
		return 0;
	}

	ioc = 1;
	ioctl (fd, FIONBIO, &ioc);
	if (connect (fd, (struct sockaddr *) &caddr, addrlen) == -1)
	{
		if (errno != EINPROGRESS)
		{
			if (type == CAN_ROSE)
			{
				struct rose_cause_struct rose_cause;

				switch (errno)
				{
				case ENETUNREACH:
					ioctl (fd, SIOCRSGCAUSE, &rose_cause);
					break;
				case EISCONN:
					rose_cause.cause = ROSE_NUMBER_BUSY;
					rose_cause.diagnostic = 0x48;
					break;
				case EINVAL:
				default:
					rose_cause.cause = ROSE_LOCAL_PROCEDURE;
					rose_cause.diagnostic = 0;
					break;
				}
				if (STREAMS_level[StreamNum] & LEVEL_linkSetup)
					BUFFERS_printBuff (StreamNum, IN, "*** Failure with %s\r", QsoTab[StreamNum].CallSign);
				printf ("*** Failure with %s\n*** %02X%02X - %s\n",
						QsoTab[StreamNum].CallSign,
						rose_cause.cause,
						rose_cause.diagnostic,
						reason (rose_cause.cause, rose_cause.diagnostic));
			}
			else
			{
				if (STREAMS_level[StreamNum] & LEVEL_linkSetup)
					BUFFERS_printBuff (StreamNum, IN, "*** Failure with %s\r", QsoTab[StreamNum].CallSign);
				printf ("*** Failure with %s\n*** %s\n",
						QsoTab[StreamNum].CallSign,
						strerror (errno));
			}
			close (fd);
			SWITCH_nSocket--;
			QsoTab[StreamNum].Type = type;
			QsoTab[StreamNum].State = FAILURE;
			QsoTab[StreamNum].fd = 0;
			QsoTab[StreamNum].Delta = TRUE;
			return 0;
		}
	}
	
	if( type == CAN_INET )
	{
		BUFFERS_printBuff(StreamNum, IN, "Connected to %s (INET)\r",
			QsoTab[StreamNum].CallSign);
	}

	QsoTab[StreamNum].Paclen = paclen;		//F5MZN 07/02/1999
	QsoTab[StreamNum].Type = type;
	QsoTab[StreamNum].State = CPROGRESS;
	QsoTab[StreamNum].fd = fd;

	return TRUE;
}

//------------------------------------------------------------------------
//------ Retourne le nombre de trames qui n'ont pas encore t ACKed -------
//------------------------------------------------------------------------
int SWITCH_getUnack (int StreamNum)
{
	int ack;

	if ((QsoTab[StreamNum].fd == 0) || (QsoTab[StreamNum].State != CONNECT))
		return 0;

	return (s_free (StreamNum) + 127) / 128;
}

//---------------------------------------------------------------------------
//------ Initialiser BPQ (si ncessaire) ------------------------------------
//---------------------------------------------------------------------------
void SWITCH_bpqSetup (int Mask, int Flag1, int Flag2)
{
	PORT *Port;

	if (Flag1 == 0)
	{
		// Interdit les connexions
		while (PortHead)
		{
			if (PortHead->Sock >= 0)
				close (PortHead->Sock);
			Port = PortHead;
			PortHead = PortHead->Next;
			free (Port);
		}
	}
}

// Donne la longueur actuelle de la queue
static int s_free (int StreamNum)
{
	struct ax25_info_struct ax25info;

	if (QsoTab[StreamNum].fd == 0)
		return 0;

	memset (&ax25info, 0, sizeof (struct ax25_info_struct));

	ioctl (QsoTab[StreamNum].fd, SIOCAX25GETINFO, &ax25info);
	return (ax25info.snd_q);
}

// Donne le status courant d'un stream
static int s_status (int StreamNum)
{
	QSO *QsoPtr = &QsoTab[StreamNum];
	int nb;
	int res = 0;
	fd_set sock_read;
	fd_set sock_write;
	fd_set sock_excep;
	struct timeval to;

	if (QsoPtr->fd <= 0)
		return (0);

	to.tv_sec = to.tv_usec = 0;
	QsoPtr->Event = 0;

	FD_ZERO (&sock_read);
	FD_ZERO (&sock_write);
	FD_ZERO (&sock_excep);

	FD_SET (QsoPtr->fd, &sock_read);
	FD_SET (QsoPtr->fd, &sock_write);
	FD_SET (QsoPtr->fd, &sock_excep);

	nb = select (QsoPtr->fd + 1, &sock_read, &sock_write, &sock_excep, &to);
	if (nb == -1)
	{
		perror ("select");
		return (res);
	}
	else if (nb == 0)
	{
		return (res);
	}
	else
	{
		if (FD_ISSET (QsoPtr->fd, &sock_read))
		{
			res |= READ_EVENT;
		}
		if (FD_ISSET (QsoPtr->fd, &sock_write))
		{
			if (QsoPtr->State == CPROGRESS)
			{
				nb = write (QsoPtr->fd, &res, 0);
				if (nb != -1)
				{
					QsoPtr->State = CONNECT;
					QsoPtr->Event = CONN_EVENT;
				}
				res |= EXCEPT_EVENT;
			}
			else
				res |= WRITE_EVENT;
		}
		if (FD_ISSET (QsoPtr->fd, &sock_excep))
		{
			/* printf ("Exception sur %d\n", QsoPtr->fd); */
			QsoPtr->Event = NO_EVENT;
			res |= EXCEPT_EVENT;
		}
	}
	return (res);
}

// Transforme une chaine de connexion INTERNET en structure sockaddr_rose
static int in_convert_call (char *desti, struct sockaddr_in *addr, char *Callsign)
{
	char *call, *command, *address, *port;

	command = strdup (desti);
	if (command)
	{

		call = strtok (command, " \t\n\r");
		address = strtok (NULL, " \t\r\n");
		port = strtok (NULL, " \t\r\n");

		if (call == NULL)
		{
			free (command);
			return -1;
		}
		strncpy (Callsign, strupr (call), 9);
		Callsign[10] = '\0';

		if (address == NULL)
		{
			free (command);
			return -1;
		}
		addr->sin_family = AF_INET;
		addr->sin_addr.s_addr = inet_addr (address);
		addr->sin_port = htons ((port) ? atoi (port) : 23);

		free (command);
	}
	return sizeof (struct sockaddr_in);
}

// Transforme une chaine de connexion NetRom en structure full_sockaddr_ax25
static int nr_convert_call (char *address, struct full_sockaddr_ax25 *addr)
{
	char buffer[100], *call, *alias;
	FILE *fp;
	int addrlen;

	call = address;

	while (*call)
	{
		if (isspace (*call))
		{
			*call = '\0';
			break;
		}
		*call = toupper (*call);
		++call;
	}

	if ((fp = fopen ("/proc/net/nr_nodes", "r")) == NULL)
	{
		fprintf (stderr, "call: NET/ROM not included in the kernel\n");
		return -1;
	}

	while (fgets (buffer, 100, fp) != NULL)
	{
		call = strtok (buffer, " \t\n\r");
		alias = strtok (NULL, " \t\n\r");

		if (strcmp (address, call) == 0 || strcmp (address, alias) == 0)
		{
			addrlen = convert_call (call, addr);
			addr->fsa_ax25.sax25_family = AF_NETROM;
			fclose (fp);
			if (addrlen == -1) 
				return -1;
			else
				return sizeof (struct sockaddr_ax25);
		}
	}

	fclose (fp);

	fprintf (stderr, "call: NET/ROM callsign or alias not found\n");

	return -1;
}

// Transforme une chaine de connexion ROSE en structure sockaddr_rose
static int rs_convert_call (char *address, struct sockaddr_rose *addr)
{
	char *command, *call, *rose;

	command = strdup (address);
	if (command)
	{

		call = strtok (command, " \t\n\r");
		rose = strtok (NULL, " \t\r\n");

		addr->srose_family = AF_ROSE;
		addr->srose_ndigis = 0;

		if (convert_call_entry (call, addr->srose_call.ax25_call) == -1)
		{
			free (command);
			return -1;
		}

		if (convert_rose_address (rose, addr->srose_addr.rose_addr) == -1)
		{
			free (command);
			return -1;
		}

		free (command);
	}
	return sizeof (struct sockaddr_rose);
}

// Transforme l'adresse ROSE en chaine
static char *rose_ntoa (char *peer, const rose_address * rsa)
{
	sprintf (peer, "%02X%02X%02X", rsa->rose_addr[2],
			 rsa->rose_addr[3],
			 rsa->rose_addr[4]);
	return peer;
}

// Transforme le DNIC ROSE en chaine
static char *rose_ntod (char *peer, const rose_address * rsa)
{
	sprintf (peer, "%02X%02X", rsa->rose_addr[0], rsa->rose_addr[1]);
	return peer;
}

// Connexion recue
static int sock_connexion (int new_fd, int StreamNum, union sockaddr_ham *addr)
{
	char User[80];
	char Node[80];
	char Dnic[80];
	char Digi[128];
	union sockaddr_ham addrham;
	unsigned int plen;
	char* pszPortName = NULL;

	/* Look for informations on the connection */
	plen = sizeof (union sockaddr_ham);

	if (getsockname (new_fd, (struct sockaddr *) &addrham, (int*) &plen) == 0)
	{
		QsoTab[StreamNum].Delta = TRUE;

		switch (addrham.axaddr.fsa_ax25.sax25_family)
		{
		case AF_AX25:			
			QsoTab[StreamNum].Type = CAN_AX25;			
			ax25_ntoa (User, &addr->axaddr.fsa_ax25.sax25_call);
			Digi[0] = '\0';
			if (addr->axaddr.fsa_ax25.sax25_ndigis)
			{
				int i;
				char c_digi[AX25_CALLSID + 1];

				strcat (Digi, " via");
				for (i = 0; i < addr->axaddr.fsa_ax25.sax25_ndigis; i++)
				{
					c_digi[0] = ' ';
					ax25_ntoa (c_digi + 1, &addr->axaddr.fsa_digipeater[i]);
					strcat (Digi, c_digi);
				}
			}
			if (QsoTab[StreamNum].State != CPROGRESS)
				QsoTab[StreamNum].State = CONNECT;
			printf ("*** Connected to %s%s (AX25)\n", User, Digi);			
			pszPortName = ax25_config_get_port(&addrham.axaddr.fsa_digipeater[0]);
			break;
			
		case AF_NETROM:
			QsoTab[StreamNum].Type = CAN_NETROM;
			ax25_ntoa (User, &addr->axaddr.fsa_ax25.sax25_call);
			ax25_ntoa (Node, &addr->axaddr.fsa_digipeater[0]);
			if (QsoTab[StreamNum].State != CPROGRESS)
				QsoTab[StreamNum].State = CONNECT;
			printf ("*** Connected to %s via %s (NROM)\n", User, Node);
			pszPortName = nr_config_get_port(&addrham.axaddr.fsa_ax25.sax25_call);
			break;
		case AF_ROSE:
			QsoTab[StreamNum].Type = CAN_ROSE;
			rose_ntod (Dnic, &addr->rsaddr.srose_addr);
			rose_ntoa (Node, &addr->rsaddr.srose_addr);
			ax25_ntoa (User, &addr->rsaddr.srose_call);
			if (QsoTab[StreamNum].State != CPROGRESS)
				QsoTab[StreamNum].State = CONNECT;
			printf ("*** Connected to %s via %s %s (ROSE)\n", User, Dnic, Node);
			pszPortName = rs_config_get_port(&addrham.rsaddr.srose_addr);
			break;
		case AF_INET:
			QsoTab[StreamNum].Type = CAN_INET;
			if (QsoTab[StreamNum].State /*!= CPROGRESS*/ == DISCONNECT )
			{
				strcpy (User, "INTERNET");
				QsoTab[StreamNum].State = WAITCALL;
				/* Demande l'acces + password */
				QsoTab[StreamNum].Echec = 5;
				BUFFERS_printBuff (StreamNum, OUT, Hello, TelnetInit, MyCall, Callsign);
				QsoTab[StreamNum].Delta = FALSE;
			}
			else
				strcpy (User, QsoTab[StreamNum].CallSign);
			printf ("*** Connected to %s (INET)\n", User);
			pszPortName = "inet";
			break;
		}
	}
	else
	{
		close (new_fd);
		SWITCH_nSocket--;
		return (0);
	}

	strcpy (QsoTab[StreamNum].CallSign, User);
	QsoTab[StreamNum].fd = new_fd;

	if( pszPortName && strlen(pszPortName) < 20 )
		strcpy(QsoTab[StreamNum].Port, pszPortName);
	else
		QsoTab[StreamNum].Port[0] = SNULL;

	return (TRUE);
}

static char *ax25_ntoa (char *peer, const ax25_address * axa)
{
	const unsigned char *cp;
	unsigned char *h = (unsigned char *) peer;
	short cnt;

	for (cnt = 0, cp = (unsigned char *) axa;; cnt++, cp++)
	{
		if (*cp == '\100')		/* Found a PAD char, skip the rest * */
			cnt = 5;
		else if (cnt < 6)		/* Callsign * */
			*h++ = *cp >> 1;
		else if (cnt == 6)
		{						/* SSID * */
			*h++ = '-';
			if ((*cp & 0x1e) > 19)
			{
				*h++ = '1';
				*h++ = ((*cp & 0x1e) >> 1) + '&';
			}
			else
			{
				*h++ = ((*cp & 0x1e) >> 1) + '0';
			}
			*h = '\0';
			return peer;
		}
	}
}

//--------------------------------------------------------------------------
//------ Quitter le ax25 kernel --------------------------------------------
//--------------------------------------------------------------------------
void SWITCH_exitPcFlex (void)
{
	int i;

	for (i = 0; i < NBSTREAMS + 1; i++)
	{
		if (QsoTab[i].fd > 0)
		{
			close (QsoTab[i].fd);
			QsoTab[i].fd = -1;
		}
	}
}

static int password_ok (char *call, char *str)
{
	FILE *fptr;
	int ok = 0;
	char ligne[80];
	char callsign[80];
	char password[80];

	fptr = fopen ("system/passwd.dat", "r");
	if (fptr == NULL)
		return (0);

	while (fgets (ligne, sizeof (ligne), fptr))
	{
		sscanf (ligne, "%s %s", callsign, password);

		if ((strcasecmp (call, callsign) == 0) && (strcasecmp (str, password) == 0))
		{
			ok = 1;
			break;
		}
	}

	fclose (fptr);
	return (ok);
}

#define MIN(a, b) (((a) < (b)) ? (a) : (b))

static int telnet_check (int StreamNum, char *buffer, int ilen)
{
	switch (QsoTab[StreamNum].State)
	{

	case WAITCALL:
		buffer[ilen] = '\0';
		TOOLS_removeNR (buffer);
		strupr (buffer);
		/* Recupere l'indicatif */
		while ((ilen) && (*buffer <= ' '))
		{
			++buffer;
		}

		ilen = strlen (buffer);

		/* Modif F5MZN : 16 oct 1999 */
		if ( ! TOOLS_isCall (buffer) )
		{
			if (--QsoTab[StreamNum].Echec <= 0)
			{
				SWITCH_discSwitch (StreamNum);
			}
			else
				BUFFERS_printBuff (StreamNum, OUT, Callsign);
			return 0;
		}

		strncpy (QsoTab[StreamNum].CallSign, buffer, MIN (20, ilen));
		QsoTab[StreamNum].CallSign[MIN (20, ilen)] = '\0';

		BUFFERS_printBuff (StreamNum, OUT, Password);
		QsoTab[StreamNum].State = WAITPASS;
		QsoTab[StreamNum].Echec = 3;
		return 0;

	case WAITPASS:
		buffer[ilen] = '\0';
		TOOLS_removeNR (buffer);
		/* Recupere l'indicatif */
		while ((ilen) && (*buffer <= ' '))
		{
			++buffer;
			--ilen;
		}

		/* Checks the password for the callsign */
		if (!password_ok (QsoTab[StreamNum].CallSign, buffer))
		{
			if (--QsoTab[StreamNum].Echec <= 0)
			{
				SWITCH_discSwitch (StreamNum);
				return (0);
			}
			else
				BUFFERS_printBuff (StreamNum, OUT, Password);
			return 0;
		}

		/* Connection accepted */
		QsoTab[StreamNum].State = CONNECT;
		QsoTab[StreamNum].Delta = TRUE;
		BUFFERS_printBuff (StreamNum, OUT, Logon);
		return 0;

	default:
		return (ilen);
	}
}

static char *reason (unsigned char cause, unsigned char diagnostic)
{
	static char *desc;

	switch (cause)
	{
	case ROSE_DTE_ORIGINATED:
		desc = "Remote Station cleared connection";
		break;
	case ROSE_NUMBER_BUSY:
		if (diagnostic == 0x48)
			desc = "Remote station is connecting to you";
		else
			desc = "Remote Station is busy";
		break;
	case ROSE_INVALID_FACILITY:
		desc = "Invalid X.25 Facility Requested";
		break;
	case ROSE_NETWORK_CONGESTION:
		desc = "Network Congestion";
		break;
	case ROSE_OUT_OF_ORDER:
		desc = "Out of Order, link unavailable";
		break;
	case ROSE_ACCESS_BARRED:
		desc = "Access Barred";
		break;
	case ROSE_NOT_OBTAINABLE:
		desc = "No Route Defined for address specified";
		break;
	case ROSE_REMOTE_PROCEDURE:
		desc = "Remote Procedure Error";
		break;
	case ROSE_LOCAL_PROCEDURE:
		desc = "Local Procedure Error";
		break;
	case ROSE_SHIP_ABSENT:
		desc = "Remote Station not responding";
		break;
	default:
		desc = "Unknown";
		break;
	}

	return desc;
}

//---------------------------------------------------------------------------
//------ SWITCH_getPaclen - Retourne le PACLEN pour ce stream ---------------
//---------------------------------------------------------------------------
int SWITCH_getPaclen(int StreamNum)
{
	return QsoTab[StreamNum].Paclen;
}

/*---------------------------------------------------------------------------
  ------ SWITCH_isConsoleConnected - Indique si la console est connectee ----
  ------ via TCP/IP ---------------------------------------------------------
  ---------------------------------------------------------------------------*/
int SWITCH_isConsoleConnected(void)
{
	return IsConsoleConnected;
}

/*----------------------------------------------------------------------------
  ------ Transmet le monitoring aux clients fbb ------------------------------
  ----------------------------------------------------------------------------*/
void SWITCH_monitor(int StreamNum, char * pszString, int nSize, int Direction)
{
	extern char STREAMS_level[MAX_STREAMS];
	extern char STREAMS_callsign[MAX_STREAMS][10];
	extern char PROTOCOL_linkType[65];
	

	static char szOldHeader[256];
	
	char szHeader[256];
	char szBuffer[1024];
	int  nChannel;
	int  nIsClulink = FALSE;
	
	if( STREAMS_callsign[StreamNum][0] == '\0' )
		SCRIPT_getLinkSetupCall(StreamNum, szBuffer);
	else
		strcpy(szBuffer, STREAMS_callsign[StreamNum]);
			
	if( Direction == 0 ) /* IN */
		sprintf(szHeader, "%s > %s:\r", szBuffer, MyCall);
	else /* OUT */
		sprintf(szHeader, "%s > %s:\r",	MyCall, szBuffer);
	
	if( STREAMS_level[StreamNum] & LEVEL_cluster || 
	    STREAMS_level[StreamNum] & LEVEL_linkSetup )
		nChannel = _MONITOR;
	else
		nChannel = _ALLCHAN;
		
	if( STREAMS_level[StreamNum] & LEVEL_cluster )
	{
		nChannel = _MONITOR;
		if( PROTOCOL_linkType[StreamNum] == CLU_LINK )
			nIsClulink = TRUE;
	}
	else
		nChannel = _ALLCHAN;
		
	if( strcmp(szOldHeader, szHeader) )
	{
		SWITCH_orbWrite (nChannel, szHeader, strlen(szHeader), 0, 0);
		strcpy(szOldHeader, szHeader);
	}

	if( nIsClulink )
	{
		int  i;
		char *ptr;
		sprintf(szBuffer, "<CLU%03d>", pszString[1]);
		ptr = strchr(szBuffer, '\0');
		for(i = 0; i < nSize; i++)	
		{
			char szByte[16];
			sprintf(szByte, "%02x", (unsigned char) pszString++);
			strupr(szByte);
			*ptr++ = ' ';
			*ptr++ = szByte[0];
			*ptr++ = szByte[1];

		}
		*ptr++ = '\r';
		*ptr++ = '\0';
		
		nSize = strlen(szBuffer);
	}
	else
	{
		strncpy(szBuffer, pszString, nSize);
		szBuffer[nSize] = '\0';
		TOOLS_N2R(szBuffer);
	}
	SWITCH_orbWrite (nChannel, szBuffer, nSize, 3, 0);
}

/*---------------------------------------------------------------------------
  ------ SWITCH_getPortName - Retourne le nom du port -----------------------
  ---------------------------------------------------------------------------*/
char* SWITCH_getPortName(int StreamNum, char* buffer)
{
//	unsigned int addr_len = sizeof (union sockaddr_ham);
//	union sockaddr_ham addr;

	if( ! QsoTab[StreamNum].fd )
		strcpy(buffer, "-");
	else if( QsoTab[StreamNum].Type == CAN_INET )
		strcpy(buffer, "inet");
	else if( QsoTab[StreamNum].fd )
		strcpy(buffer, QsoTab[StreamNum].Port);	
#if 0
	else if(getpeername (QsoTab[StreamNum].fd, (struct sockaddr *) &addr, &addr_len) == 0)
	{
		switch (addr.axaddr.fsa_ax25.sax25_family)
		{
		case AF_AX25:			
			strcpy(buffer, ax25_config_get_port(&addr.axaddr.fsa_ax25.sax25_call));
			break;
		case AF_NETROM:			
			strcpy(buffer, nr_config_get_port(&addr.axaddr.fsa_ax25.sax25_call));
			break;
		case AF_ROSE:			
			strcpy(buffer, rs_config_get_port(&addr.rsaddr.srose_addr));
			break;	
		default :
			strcpy(buffer, "?ERR");	
			break;
		}
	}
#endif
	return buffer;
}
